#ifndef ICLIENT_HH
#define ICLIENT_HH

#include <iostream>
#include <sstream>

#include <errno.h>
#include <sys/types.h>

//Define necessary to enable getaddrinfo
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x602
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <Ws2tcpip.h>

#include <imsg.hh>
#include <message.hh>
#include <msgConstants.hh>

class iclient {

private:
	SOCKET sock;

public:
	enum received{STAT_ERR, STAT_EMPT, STAT_RECV};

	iclient(): sock(-1) {}
	
	~iclient(){
		disconnect();
	}

	bool connect(const std::string& serverName, int portNumber){
		WSADATA WSAData;
		int res = WSAStartup(MAKEWORD(2,0), &WSAData);
		if(res != 0){
			std::cerr<<"ERROR: Can not initialise winSocket:"<<std::endl;
			switch(res){
				case WSASYSNOTREADY:
					std::cerr<<"The underlying network subsystem is not ready for network communication."<<std::endl;
					break;
				case WSAVERNOTSUPPORTED:
					std::cerr<<"The version of Windows Sockets support requested is not provided by this particular Windows Sockets implementation."<<std::endl;
					break;
				case WSAEINPROGRESS:
					std::cerr<<"A blocking Windows Sockets 1.1 operation is in progress."<<std::endl;
					break;
				case WSAEPROCLIM:
					std::cerr<<"A limit on the number of tasks supported by the Windows Sockets implementation has been reached."<<std::endl;
					break;
				case WSAEFAULT:
					std::cerr<<"The lpWSAData parameter is not a valid pointer."<<std::endl;
					break;
				default:
					std::cerr<<"Unknown error: "<<res<<std::endl;
			}
			return false;
		}
		
		if(isConnected()){
			//If already coonected, drop the last coonection
			disconnect();
		}
		
		//Set up the socket
		struct addrinfo * address, *current;
		struct addrinfo hints;
		memset(&hints, 0, sizeof(struct addrinfo)); //Create empty hint
		hints.ai_family=AF_UNSPEC; //For both IPv4 and IPv6
		hints.ai_socktype=SOCK_STREAM;
		std::stringstream port;
		port<<portNumber;
		int result=getaddrinfo(serverName.c_str(), port.str().c_str(), &hints, &address);
		if(result!=0){
			std::cerr<<"Error while creating the address :"<<std::endl;
			std::cerr<<gai_strerror(result)<<std::endl;
			return false;
		}
		
		for (current=address; current != NULL; current=current->ai_next) {
			sock=socket(current->ai_family, current->ai_socktype, current->ai_protocol);
			if(sock==INVALID_SOCKET){
				continue;
			}
			
			//Allow multiple connexions
			int reuse_addr=1;
			if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*) &reuse_addr, sizeof(reuse_addr))!=0){
				disconnect();
				continue;
			}
			
			if(::connect(sock, current->ai_addr, current->ai_addrlen) == 0) //Connection successful
				break;
			
			disconnect();
		}
		
		freeaddrinfo(address);
		
		if(current==NULL || sock==INVALID_SOCKET || result==-1){
			std::cerr<<"Error while creating socket :"<<std::endl;
			wchar_t* message=NULL;
			FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
			   NULL, WSAGetLastError(),
			   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			   message, 0, NULL);
			std::cerr<<message<<std::endl;
			free(message);
			return false;
		}

		return true;
	}

	void subscribe(const std::string& name) const{
		message post;
		post.encode(DEFAULT_SERVER_NAME, name.c_str());
		send(post);
	}

	bool isConnected() const{
		if(sock==INVALID_SOCKET) return false;
		//Check the socket state
		int error;
		socklen_t len=sizeof(error);
		int retval=getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&error, &len);
		return(retval==0);
		//If retval == 0 the socket is up, otherwise it is disconnected (it allow to detect only some errors)
	}

	void disconnect() const{
		closesocket(sock);
		WSACleanup();
	}

	bool checkReadSocket(bool blocking) const{
		if(!isConnected()){
			std::cerr<<"Server disconnected"<<std::endl;
			disconnect();
			return false;
		}
		int timeout=0; //Non blocking
		if(blocking){
			timeout=-1; //Wait forever
		}
		return(checkReadSocketTimeout(timeout));
	}
	
	//ms=-1 - blocking waiting
	//ms=0 - non blocking waiting
	//ms>0 - wait with timeout
	bool checkReadSocketTimeout(int ms) const{
		if(!isConnected()){
			std::cerr<<"Server disconnected"<<std::endl;
			disconnect();
			return false;
		}
		fd_set fd, errors;
		FD_ZERO(&fd);
		FD_SET(sock, &fd);
		FD_ZERO(&errors);
		FD_SET(sock, &errors);
		
		int result;
		if(ms>=0){
			timeval time;
			time.tv_sec=((long)ms)/1000;
			time.tv_usec=(((long)ms)%1000)*1000;
			
			result=select(sock+1, &fd, NULL, &errors, &time);
		}else{ //Blocking socket
			result=select(sock+1, &fd, NULL, &errors, NULL);
		}
				
		if(FD_ISSET(sock, &errors)){
			//TODO Retrieve error code for the socket
			// int error;
			// socklen_t len=sizeof(error);
			// int retval=getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&error, &len);
			disconnect();
			return false;			
		}
		if(result==0){ //Timed out
			return false;
		}
		if(result==SOCKET_ERROR){
			std::cerr<<"Error on the socket while waiting for a message"<<std::endl;
			wchar_t* message=NULL;
			FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
			   NULL, WSAGetLastError(),
			   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			   message, 0, NULL);
			std::cerr<<message<<std::endl;
			disconnect();
			return false;
		}
		return(FD_ISSET(sock, &fd));
	}
	
	bool send(const message& data) const{
		if(!isConnected()){
			std::cerr<<"Server disconnected"<<std::endl;
			disconnect();
			return false;
		}
		imsg msg;
		if(!msg.encode(data.data()))
			return false;
		ssize_t n;
		size_t tot = 0;
		size_t bleft = msg.length();		
		while(tot < msg.length()) {
			n = ::send(sock, msg.data() + tot, bleft, 0);
			if(n < 1){
				std::cerr<<"Error while sending the message"<<std::endl;
				wchar_t* message=NULL;
				FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
					NULL, WSAGetLastError(),
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					message, 0, NULL);
				std::cerr<<message<<std::endl;
				free(message);
				disconnect();
				break;
			}
			tot += n;
			bleft -= n;
		}
		return (tot == msg.length());
	}

	//Ugly hack to a missing definition in Windows' socket
	#ifndef MSG_WAITALL
	#define MSG_WAITALL 0x8
	#endif
	std::pair<received, message> receive() const{
		message result;
		imsg msg;
		ssize_t res;
		char header[msg.hsize()];
		errno=0;

		if(!isConnected()){
			std::cerr<<"Server disconnected"<<std::endl;
			disconnect();
			return std::make_pair(STAT_ERR, result);
		}
		
		if(!checkReadSocket(false)){
			std::cerr<<"EMPTY SOCKET"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}

		res = ::recv(sock, header, msg.hsize(), 0); //0 = MSG_DONTWAIT
		if(res==0){
			std::cerr<<"Server disconnected"<<std::endl;
			disconnect();
			return std::make_pair(STAT_ERR, result);
		} else if(res==SOCKET_ERROR){
			std::cerr<<"Error on the socket while receiving a message :"<<std::endl;
			wchar_t* message=NULL;
			FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
			   NULL, WSAGetLastError(),
			   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			   message, 0, NULL);
			std::cerr<<message<<std::endl;
			free(message);
			disconnect();
			return std::make_pair(STAT_ERR, result);
		}
		
		if(res != static_cast<ssize_t>(msg.hsize())){
			std::cerr<<"Error while reading socket, header does not have the correct length"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeHeader(header);

		char body[msg.blength()];
		res = ::recv(sock, body, msg.blength(), MSG_WAITALL);
		if(res != static_cast<ssize_t>(msg.blength())){
			std::cerr<<"Error while reading socket, body does not have the correct length"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeBody(body);

		result.decode(msg.body());
		return std::make_pair(STAT_RECV, result);
	}
};

#endif
